'' ============================================================
'' 3D Simulated Solid Body Particle System
''
'' Copyright (C) 2006 - Jeff Marshall (coder@execulink.com)
''
'' This software is provided 'as-is', without any express or implied
'' warranty. In no event will the authors be held liable for any damages
'' arising from the use of this software.
''	
'' Permission is granted to anyone to use this software for any purpose,
'' including commercial applications, and to alter it and redistribute it
'' freely, subject to the following restrictions:
''	
'' 1. The origin of this software must not be misrepresented; you must not
'' claim that you wrote the original software. If you use this software in
'' a product, an acknowledgment in the product documentation would be
'' appreciated but is not required.
''	
'' 2. Altered source versions must be plainly marked as such, and must not
'' be misrepresented as being the original software.
''	
'' 3. This notice may not be removed or altered from any source
'' distribution.
''
'' ============================================================
''
'' This particle system is not based on any real world model but
'' it does have some theory.  The math is simplified by having
'' all "atoms" of the same size and mass.  (There is no mass in
'' any of the calculations)  Also, certain physical effects are
'' simulated using factors and conditions to change the state of
'' the system.
''
'' The solid bodies are made up of "atoms". 
''
'' * Atoms have a spacing and a size (ATOMSPACE, ATOMSIZE)
'' * Atoms are affected by gravity (XGRAVITY, YGRAVITY, ZGRAVITY)
'' * Atoms can bounce off the sides and bottom of the screen
''  and when they do, lose some momentum due to the
''  collision (BOUNDDECAY)
'' * Atoms have a max speed (MAXSPEED) to limit values and
''  make the simulation work better.
''
'' "Bonds" represent the relationship between two atoms.
'' When a bond is first created a distance between two atoms
'' is set to indicate how far the two atoms should be from
'' each other.  Too far or too close and the system will move
'' the atoms closer or further apart.  Let's call this the
'' balanced distance.
''
'' * The rate at which two atoms return to the balanced
''   distance is affected by stiffness (STIFFNESS).
'' * if two atoms are moving away from each other and the
''   instantaneous velocity is high enough the bond is
''   broken. (BREAKTHRESHOLD)
'' * if the magnitude of the relative velocity between two
''   atoms is high enough (DEFORMTHRESHOLD) the bond can be
''   deformed.  The amount of deformation is controlled
''   by a factor. (DEFORMFACTOR)
'' * If a collision between two atoms is detected and the
''   rate of approach is high enough, a new bond may be formed
''   (BONDTHRESHOLD)
''
'' ============================================================

Option Explicit
const False = 0, True = Not False

#include "vector3.bi"
#include "atomseng3.bi"

dim shared as integer MAXATOMS = 16
dim shared as integer MAXBONDS = 16

dim shared as integer NATOMS = 0
dim shared as integer NBONDS = 0

dim shared as single ATOMSIZE = 4
dim shared as single ATOMSPACE = 10
dim shared as single MAXSPEED = 5

dim shared as single ZGRAVITY = -.004
dim shared as single YGRAVITY = 0
dim shared as single XGRAVITY = 0
dim shared as single BOUNDDECAY = .99
dim shared as single DEFORMFACTOR = .4
dim shared as single DEFORMTHRESHOLD = 1
dim shared as single BREAKTHRESHOLD = 2
dim shared as single STIFFNESS = 0.5
dim shared as single BONDTHRESHOLD = 0.5

dim shared as integer SHOWATOMS = 1
dim shared as integer SHOWBONDS = 1

redim shared a(1 TO MAXATOMS) as Atom
redim shared b(1 TO MAXBONDS) as Bond

'' ============================================================
'' ATOM Primitives
'' ============================================================

sub ClearAllAtoms
	NATOMS = 0
	NBONDS = 0									
end sub

function GetFreeAtom() as integer
  dim i as integer
  function = 0
  for i = 1 TO NATOMS
    if a(i).ex = FALSE then
			exit for
		end if
	next i
	if i > NATOMS then
		NATOMS += 1
		i = NATOMS
		if NATOMS > MAXATOMS then
			MAXATOMS shl= 1
			redim preserve a(1 to MAXATOMS) as Atom
		end if
	end if
	a(i).ex = FALSE
	function = i
end function

function GetFreeBond() as integer
  dim i as integer
  function = 0
  for i = 1 TO NBONDS
    if b(i).ex = FALSE then
			exit for
		end if
	next i
	if i > NBONDS then
		NBONDS += 1
		i = NBONDS
		if NBONDS > MAXBONDS then
			MAXBONDS shl= 1
			redim preserve b(1 to MAXBONDS) as Bond
		end if
	end if
	b(i).ex = FALSE
	function = i
end function

sub SetAtom (byval idx as integer, byval x1 as single, byval y1 as single, byval z1 as single, byval dx as single, byval dy as single, byval dz as single, byval c as integer)
  a(idx).ex = TRUE
  a(idx).p.x = x1
  a(idx).p.y = y1
  a(idx).p.z = z1
  a(idx).v.x = dx
  a(idx).v.y = dy
  a(idx).v.z = dz
  a(idx).flg = FALSE
  a(idx).c = c
	a(idx).clr.x = (c shr 16 and 255) / 255
	a(idx).clr.y = (c shr 8 and 255) / 255
	a(idx).clr.z = (c and 255) / 255
end sub

sub SetBond (byval idx as integer, byval a1 as integer, byval a2 as integer, byval bal as single, byval ten as single, byval c as integer)
  b(idx).ex = TRUE
  b(idx).a1 = a1
  b(idx).a2 = a2
  b(idx).bal = bal
  b(idx).ten = ten
  b(idx).c = c
	b(idx).clr.x = (c shr 16 and 255) / 255
	b(idx).clr.y = (c shr 8 and 255) / 255
	b(idx).clr.z = (c and 255) / 255
end sub

sub RotateAtoms (byval x as single, byval y as single, byval z as single, byval rx as single, byval ry as single, byval rz as single)
  dim v as vector, i as integer
  for i = 1 to NATOMS
    if a(i).ex = TRUE then
    if a(i).flg = FALSE then
      VSubComp v, a(i).p, x, y, z
      VRot2Vx v, rx
      VRot2Vy v, ry
      VRot2Vz v, rz
      VAddComp a(i).p, v, x, y, z

      a(i).flg = TRUE
    end if
    end if
  next i
end sub

function FindAtom(byval x as single, byval y as single, byval z as single) as integer
  dim i as integer
	function = 0
  for i = 1 TO NATOMS
    if a(i).ex = TRUE then
      if Abs(x - a(i).p.x) < ATOMSPACE then
        if Abs(y - a(i).p.y) < ATOMSPACE then
	        if Abs(z - a(i).p.z) < ATOMSPACE then
						function = i
						exit for
					end if
				end if
      end if
    End if
  next i
end function

function IsBond(aidx as integer, bidx as integer) as integer
	dim i as integer
  IsBond = TRUE
  if aidx <> bidx then
    FOR i = 1 TO MAXBONDS
      if B(i).ex = TRUE then
        if B(i).a1 = aidx and B(i).a2 = bidx then
          Exit function
        end if
        if B(i).a2 = aidx and B(i).a1 = bidx then
          Exit function
        end if
      end if
    NEXT i
    IsBond = FALSE
  end if
end function

'' ============================================================
'' "PHYSICS" like functions for updating state
'' ============================================================

sub ApplyBounds (byval i as integer)

  if a(i).p.z < 0 then
    a(i).p.z = -a(i).p.z
    'a(i).v.x = -a(i).v.x * BOUNDDECAY
    'a(i).v.y = -a(i).v.y * BOUNDDECAY
    a(i).v.z = -a(i).v.z * BOUNDDECAY
  end if
 
  if a(i).p.x > 320 then
    a(i).p.x = 320 - (a(i).p.x - 320)
    a(i).v.x = -a(i).v.x * BOUNDDECAY
  end if

  if a(i).p.x < -320 then
    a(i).p.x = -320 + (-a(i).p.x - 320)
    a(i).v.x = -a(i).v.x * BOUNDDECAY
  end if

  if a(i).p.y > 320 then
    a(i).p.y = 320 - (a(i).p.y - 320)
    a(i).v.y = -a(i).v.y * BOUNDDECAY
  end if

  if a(i).p.y < -320 then
    a(i).p.y = -320 + (-a(i).p.y - 320)
    a(i).v.y = -a(i).v.y * BOUNDDECAY
  end if

end sub

sub ApplyCollision (byval i as integer)
 
  dim cij as vector
  dim cji as vector
  dim rij as vector
  dim rji as vector
	dim as integer j, bidx
	dim as single d, k, k1, k2
 
  ' Check Atom i to see if it has hit any other

  for j = 1 TO NATOMS
    if j <> i then
			if a(j).ex = TRUE then
     
				VSub cij, a(j).p, a(i).p

				if cij.x + cij.y  + cij.z < ATOMSIZE * 4 then

					VNeg cji, cij
					d = VMag(cij)
					if d < ATOMSIZE * 2 then
						' We have a collision
						' Is I moving towards J

						'' k = (cij.x * cij.x + cij.y * cij.y + cij.z * cij.z)
						k = VDot(cij, cij)
       
						' Project i.v on cij => rij
						'' k1 = (a(i).v.x * cij.x + a(i).v.y * cij.y) / k
						k1 = VDot(a(i).v, cij) / k

						if k1 >= 0 then
							if k1 > BONDTHRESHOLD then
								If IsBond(i, j) = FALSE then
									bidx = GetFreeBond()
									SetBond bidx, i, j, ATOMSPACE, ATOMSPACE, a(i).c
								End if
							end if

							VScale rij, cij, k1
       
							' Project j.v on cji => rji
							''k2 = (a(j).v.x * cji.x + a(j).v.y * cji.y) / k
							k2 = VDot(a(j).v, cji) / k
							VScale rji, cji, k2

							' Apply Results
							VAdd2V a(i).v, rji
							VSub2V a(i).v, rij
       
							VAdd2V a(j).v, rij
							VSub2V a(j).v, rji
						end if
					end if
				end if
			end if
    end if
  next j
end sub

sub ApplyTension (byval i as integer)
 
  dim v as vector
	dim as integer a1, a2
	dim as single diff, d
 
  a1 = b(i).a1
  a2 = b(i).a2

  VSub v, a(a1).p, a(a2).p

  d = VMag(v)

  if d / ATOMSPACE > BREAKTHRESHOLD then
    b(i).ex = FALSE
    EXIT sub
  end if
 
  diff = (d - b(i).bal) / 2 * STIFFNESS
 
  VScale2V v, diff / b(i).bal
  if ABS(d / ATOMSPACE) > DEFORMTHRESHOLD then
    b(i).bal = b(i).bal + diff * DEFORMFACTOR
  end if
  
  VAdd2V a(a2).v, v
 
  VNeg2V v
  VAdd2V a(a1).v, v

end sub

sub UpdateAtoms ()
	dim i as integer

  for i = 1 TO NBONDS
    if b(i).ex = TRUE then
      ApplyTension i
    end if
  next i
 
  for i = 1 TO NATOMS
    if a(i).ex = TRUE then
      VAddComp2V a(i).v, XGRAVITY, YGRAVITY, ZGRAVITY  ' Apply Gravity
     
      ApplyBounds i
     
      ApplyCollision i
     
      VLimit2V a(i).v, MAXSPEED               ' Limit Speed
      VAdd2V a(i).p, a(i).v           ' Apply Movement
      
    end if
  next i
end sub

'' ============================================================
'' SOLIDBODY CREATION
'' ============================================================

sub CreateTriBlock(byval x1 as single, byval y1 as single, byval z1 as single, byval siz as integer, byval c as integer, byval rx as single = 0, byval ry as single = 0, byval rz as single = 0)
  dim as integer x, y, z, bidx
	dim as single xspc, yspc, zspc, bal, ten, t1, t2
	xspc = ATOMSPACE
  yspc = SQR(0.75 * ATOMSPACE * ATOMSPACE)
	t1 = yspc / 2
	t2 = SQR(ATOMSPACE * ATOMSPACE / 4 + t1 * t1)
	zspc = SQR(ATOMSPACE * ATOMSPACE - t2 * t2)
  bal = ATOMSPACE
  ten = ATOMSPACE
  redim aidx(0 TO siz - 1, 0 TO siz - 1, 0 TO siz - 1)
  ' Create Atoms
	for z = 0 to siz - 1
		for x = 0 to (siz - z) - 1
			for y = 0 to (siz - z) - x - 1
				aidx(x, y, z) = GetFreeAtom
				SetAtom aidx(x, y, z), x1 + (x + (y + z) / 2) * xspc, y1 + (y + z / 2) * yspc, z1 + z * zspc, 0, 0, 0, c
			next y
		next x
	next z
  ' Create Diagonal Bonds
	for z = 0 to siz - 2
		for x = 0 to (siz - z) - 2
			for y = 0 to (siz - z) - x - 2
				SetBond GetFreeBond, aidx(x, y, z), aidx(x, y + 1, z), ATOMSPACE, ten, c
				SetBond GetFreeBond, aidx(x, y, z), aidx(x + 1, y, z), ATOMSPACE, ten, c
				SetBond GetFreeBond, aidx(x + 1, y, z), aidx(x, y + 1, z), ATOMSPACE, ten, c
				SetBond GetFreeBond, aidx(x, y, z), aidx(x, y, z + 1), ATOMSPACE, ten, c
				SetBond GetFreeBond, aidx(x, y + 1, z), aidx(x, y, z + 1), ATOMSPACE, ten, c
				SetBond GetFreeBond, aidx(x + 1, y, z), aidx(x, y, z + 1), ATOMSPACE, ten, c
			next y
		next x
	next z
	RotateAtoms x1, y1, z1, rx, ry, rz
end sub
sub CreateBlock (byval x1 as single, byval y1 as single, byval z1 as single, byval xsiz as integer, byval ysiz as integer, byval zsiz as integer, byval c as integer, byval rx as single = 0, byval ry as single = 0, byval rz as single = 0)
	dim as integer x, y, z, bidx
	dim as single xspc, yspc, zspc, bal, ten
  xspc = ATOMSPACE
  yspc = ATOMSPACE
	zspc = ATOMSPACE
  bal = ATOMSPACE
  ten = ATOMSPACE
  redim aidx(0 TO xsiz - 1, 0 TO ysiz - 1, 0 to zsiz - 1) as integer
  ' Create Atoms
  for z = 0 to zsiz - 1
		for x = 0 TO xsiz - 1
			for y = 0 TO ysiz - 1
				aidx(x, y, z) = GetFreeAtom
				SetAtom aidx(x, y, z), x1 + x * xspc, y1 + y * yspc, z1 + z * zspc, 0, 0, 0, c
			next y
		next x
	next z
	' Create Bonds
	for z = 0 to zsiz - 1
		for y = 0 to ysiz - 1
			for x = 0 to xsiz - 1
				if x < xsiz - 1 then SetBond GetFreeBond, aidx(x, y, z), aidx(x + 1, y, z), bal, ten, c
				if y < ysiz - 1 then SetBond GetFreeBond, aidx(x, y, z), aidx(x, y + 1, z), bal, ten, c
				if z < zsiz - 1 then SetBond GetFreeBond, aidx(x, y, z), aidx(x, y, z + 1), bal, ten, c
				if x < xsiz - 1 and y < ysiz - 1 then 
					SetBond GetFreeBond, aidx(x, y, z), aidx(x + 1, y + 1, z), SQR(bal * bal * 2), ten, c
					SetBond GetFreeBond, aidx(x + 1, y, z), aidx(x, y + 1, z), SQR(bal * bal * 2), ten, c
				end if
				if x < xsiz - 1 and z < zsiz - 1 then 
					SetBond GetFreeBond, aidx(x, y, z), aidx(x + 1, y, z + 1), SQR(bal * bal * 2), ten, c
					SetBond GetFreeBond, aidx(x + 1, y, z), aidx(x, y, z + 1), SQR(bal * bal * 2), ten, c
				end if
				if y < ysiz - 1 and z < zsiz - 1 then 
					SetBond GetFreeBond, aidx(x, y, z), aidx(x, y + 1, z + 1), SQR(bal * bal * 2), ten, c
					SetBond GetFreeBond, aidx(x, y + 1, z), aidx(x, y, z + 1), SQR(bal * bal * 2), ten, c
				end if
				'if x < xsiz - 1 and y < ysiz - 1 and z < zsiz - 1 then 
				'	SetBond GetFreeBond, aidx(x, y, z), aidx(x + 1, y + 1, z + 1), ???, ten, c
				'end if
			next x
		next y
	next z
	RotateAtoms x1, y1, z1, rx, ry, rz
end sub

private sub ParseArgs(x as string, w() as string, n as integer, m as integer)
	dim i as integer, j as integer
	n = 0
	i = 1
	while n < m
		j = instr(i, x, ",")
		if j = 0 then
			n = n + 1
			w(n) = mid(x,i)
			exit while
		else
			n = n + 1
			w(n) = mid(x,i,j-i)
			i = j + 1
		end if
	wend
end sub

Sub LoadSettings(strFilename as string)
dim x as string, i as integer, v as string, w as string, t(1 to 10) as string, n as integer
Open strFilename for input as #1
while eof(1) = 0
	line input #1, x
	if x > "" and left(x,1) <> "'" and left(x,1) <> "#" and left(x,1) <> ";" then
		i = instr(x,"=")
		if i > 0 then
			v = ltrim(rtrim(ucase(left(x,i-1))))
			w = ltrim(rtrim(mid(x,i+1)))
			select case v
      case "ATOMSIZE"          : ATOMSIZE = val(w)
      case "ATOMSPACE"         : ATOMSPACE = val(w)
      case "MAXSPEED"					 : MAXSPEED = val(w)
      case "ZGRAVITY"					 : ZGRAVITY = val(w)
      case "YGRAVITY"					 : YGRAVITY = val(w)
      case "XGRAVITY"					 : XGRAVITY = val(w)
      case "BOUNDDECAY"				 : BOUNDDECAY = val(w)
      case "DEFORMFACTOR"			 : DEFORMFACTOR = val(w)
      case "DEFORMTHRESHOLD"	 : DEFORMTHRESHOLD = val(w)
      case "BREAKTHRESHOLD"		 : BREAKTHRESHOLD = val(w)
      case "STIFFNESS"				 : STIFFNESS = val(w)
      case "BONDTHRESHOLD"		 : BONDTHRESHOLD = val(w)
      case "SHOWATOMS"				 : SHOWATOMS = val(w)
      case "SHOWBONDS"				 : SHOWBONDS = val(w)
			case "BOX"
				ParseArgs w, t(), n, 10
				if n = 10 then
					CreateBlock val(t(1)), val(t(2)), val(t(3)), val(t(4)), val(t(5)), val(t(6)), val(t(7)), val(t(8)), val(t(9)), val(t(10))
				end if
			case "TRIANGLE"
				ParseArgs w, t(), n, 8
				if n = 8 then
					CreateTriBlock val(t(1)), val(t(2)), val(t(3)), val(t(4)), val(t(5)), val(t(6)), val(t(7)), val(t(8))
				end if
			end select
		end if
	end if
wend
close #1
End Sub
